/*

Module		:	scr241_main.c

Author		:	SCM Microsystems

Development History :

	Date		Version No.		Changes
	_______________________________________________
	2 Feb, 2004		2.0.1			Initial Versions
	30 Mar, 2004		3.0.1			Added support for suspend, resume
									under event handler.
	22 Nov,2004		3.1.1			TTP issues 1359,1360

	
*/

#include "includes.h"

#define CS_CHECK(fn, args...) \
while ((last_ret=CardServices(last_fn=(fn),args))!=0) goto cs_failed

#define CFG_CHECK(fn, args...) \
if (CardServices(fn, args) != 0) goto next_entry

DECLARE_TASK_QUEUE(tq_scr241);

dev_link_t 			*dev_list 			= NULL;
dev_info_t 			dev_info 			= "scr24x_cs";
int				free_ports 			= 0;
u_int 				irq_mask 			= 0xdeb8;
int 				irq_list[4]	 		= { -1 };
DEVICE_EXTENSION 	*pDeviceExtension;
int 				minorn 				= 0;
int 				majorn 				= 0;
BOOL				gWaitEvent			= FALSE;

static void Driver_Config( dev_link_t *link );

struct file_operations SCR241_fops =
{
	read: 	IFDHRead,
	ioctl: 	IFDHIoctl,
	write: 	IFDHWrite,
	open: 	IFDHOpen,
	release:IFDHClose,
};

static int __init init_driver( void )
{
	int nRetVal=1;
	DebugLog ("<1> Calling Init Driver of SCR241 Module\n");
	nRetVal = register_pccard_driver(&dev_info, &Driver_Attach, &Driver_Detach);
	DebugLog ("<1> Calling Init Driver of SCR241 Module EXIT %d\n", nRetVal );
	return nRetVal;
}

static void __exit exit_driver(void)
{
	unregister_pccard_driver(&dev_info);

	while (dev_list != NULL)
	{
		del_timer(&dev_list->release);

		if (dev_list->state & DEV_CONFIG)
		{
			Driver_Release ((u_long)dev_list);
		}

		Driver_Detach(dev_list);
	}
	DebugLog ("<1> Calling Exit Driver of SCR241 Module\n");
}

dev_link_t *Driver_Attach( void )
{
	dev_link_t 		*pDeviceLink;
	client_reg_t 	client_reg;
	int 		nInstanceCount =0; 
	int 			ncsRetVal,
					nCount;

	DebugLog ("<1> Calling Driver_Attach of SCR241 Module\n");
	// memories allocation
	pDeviceLink = ( dev_link_t *)kmalloc( sizeof(dev_link_t), GFP_KERNEL );
	if( NULL == pDeviceLink )
	{
		return NULL;
	}
	memset ( pDeviceLink , 0, sizeof( dev_link_t ) );

	pDeviceExtension = (DEVICE_EXTENSION *)
						kmalloc( sizeof(DEVICE_EXTENSION),
						GFP_KERNEL );

	if( NULL == pDeviceExtension )
	{
		kfree ( pDeviceLink );
		return NULL;
	}
	memset( pDeviceExtension , 0 , sizeof( DEVICE_EXTENSION ) );

	pDeviceExtension->pucDeviceName = (PUCHAR )
						kmalloc( sizeof(UCHAR) * 32 ,
						GFP_KERNEL );

	if( NULL == pDeviceExtension->pucDeviceName )
	{
		kfree ( pDeviceExtension );
		kfree ( pDeviceLink );
		return NULL;
	}
	// End of Allocation of memories
	// setting default values to device extension structure.
	pDeviceExtension->nInstanceCount 				= 0;
	pDeviceExtension->nMajorNumber 					= 0;
	pDeviceExtension->nInstanceCount 				= nInstanceCount;
	pDeviceExtension->bBootRomPresent				= FALSE;
	pDeviceExtension->bIsPolledMode 				= FALSE;
	
	// pcsc daemon does not issue a power request to update
	// card status on system resume frmo standby/suspend. 
	// hence, a card absent/present is simuated from the
	// kernel module.
	pDeviceExtension->bIsSystemResume				= FALSE;

	//Scheduler Queue for calling cardstatus function becos
	// we cannot sleep in the interrupt handler and card status
	// function contains interruptible_wait in its loop
	pDeviceExtension->tqsCardStatusCheck.routine 	= tqGetCardStatus;
	pDeviceExtension->tqsCardStatusCheck.data 		= pDeviceExtension;
	pDeviceExtension->tqsCardStatusCheck.sync 		= 0;
	init_waitqueue_head( &pDeviceExtension->tqsEventWait );

	sprintf(pDeviceExtension->pucDeviceName,"SCR24x%d",nInstanceCount++);
	DebugLog (
		"<1> Driver_Attach of SCR241 Module:: [%s]\n",
		pDeviceExtension->pucDeviceName);

	// We specify 0 in the major number location so that the allocation is
	// done dynamically and if the registration is successful, the major
	// number is returned.
	pDeviceExtension->nMajorNumber = register_chrdev(
										0,
										pDeviceExtension->pucDeviceName,
										&SCR241_fops);

	if( pDeviceExtension->nMajorNumber < 1)
	{
		DebugLog( "<1> Cannot retrieve Major Number dynamically \n");
	}

	// Initialize the dev_link_t structure
	pDeviceLink->release.function = &Driver_Release;
	pDeviceLink->release.data = (u_long)pDeviceLink;

	// Interrupt setup
	pDeviceLink->irq.Attributes = IRQ_TYPE_DYNAMIC_SHARING | IRQ_HANDLE_PRESENT;
	pDeviceLink->irq.IRQInfo1 = IRQ_INFO2_VALID | IRQ_LEVEL_ID;
	if (irq_list[0] == -1)
	{
		pDeviceLink->irq.IRQInfo2 = irq_mask;
	}
	else
	{
		for (nCount = 0; nCount < 4; nCount++)
		{
			pDeviceLink->irq.IRQInfo2 |= 1 << irq_list[nCount];
		}
	}
	pDeviceLink->irq.Handler = &Driver_IrqHandler;
	pDeviceLink->irq.Instance = pDeviceExtension;


	//General socket configuration defaults can go here.  In this
	//client, we assume very little, and rely on the CIS for almost
	//everything.  In most clients, many details (i.e., number, sizes,
	//and attributes of IO windows) are fixed by the nature of the
	//device, and can be hard-wired here.

	pDeviceLink->conf.Attributes = CONF_ENABLE_IRQ;
	pDeviceLink->conf.Vcc = 50;
	pDeviceLink->conf.IntType = INT_MEMORY_AND_IO;

	/* Register with Card Services */
	pDeviceLink->next = dev_list;
	dev_list = pDeviceLink;

	client_reg.dev_info = &dev_info;
	client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
	client_reg.EventMask =
		CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
		CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
		CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
	client_reg.event_handler = &Event_Handler;
	client_reg.Version = 0x0100;
	client_reg.event_callback_args.client_data = pDeviceLink;
	ncsRetVal = CardServices(RegisterClient, &pDeviceLink->handle, &client_reg);

	if (ncsRetVal != CS_SUCCESS)
	{
		DebugLog ( \
			"<1> Driver_Attach SCR241 Module FAILED!!!:: [%s]\n", \
			pDeviceExtension->pucDeviceName);
		cs_error(pDeviceLink->handle, RegisterClient, ncsRetVal);
		Driver_Detach(pDeviceLink);
		return NULL;
	}
	pDeviceExtension->DeviceLink = pDeviceLink;

	pDeviceExtension->CardTracker.pDeviceExtension = pDeviceExtension;

	return pDeviceLink;
}

void Driver_Detach(dev_link_t *pDeviceLink)
{
	/*======================================================================

	This deletes a driver "instance".  The device is de-registered
	with Card Services.  If it has been released, all local data
	structures are freed.  Otherwise, the structures will be freed
	when the device is released.

	======================================================================*/
	dev_link_t **dpDeviceLink;

	DebugLog( "<1> SCR241::Driver_Detach Entry \n");

	if( NULL != pDeviceExtension )
	{
		wake_up_interruptible_sync( &pDeviceExtension->tqsEventWait ); 
		flush_scheduled_tasks();
	}
	
	/* Locate device structure */
	for (dpDeviceLink = &dev_list;
	*dpDeviceLink;
	dpDeviceLink = &(*dpDeviceLink)->next)
	{
		if ( *dpDeviceLink == pDeviceLink )
		{
			break;
		}
	}
	if (*dpDeviceLink == NULL)
	{
		return;
	}

	if (pDeviceLink->state & DEV_CONFIG)
	{
		pDeviceLink->state |= DEV_STALE_LINK;
		return;
	}

	/* Kill the card tracking thread */
	if( TRUE == pDeviceExtension->bIsPolledMode )
	{
		DebugLog(
		"<1> SCR241::Stopping the Polled Mode Card Tracking Thread.\n");
		pDeviceExtension->bIsPolledMode = FALSE;
		pDeviceExtension->bIsIrqSet = FALSE;
		stop_kthread( &pDeviceExtension->CardTracker );
	}

	/* Break the link with Card Services */
	if (pDeviceLink->handle)
	{
		CardServices(DeregisterClient, pDeviceLink->handle);
	}

	/* Unlink device structure, and free it */
	*dpDeviceLink = pDeviceLink->next;
	/* This points to the parent local_info_t struct */

	unregister_chrdev(
					pDeviceExtension->nMajorNumber,
					pDeviceExtension->pucDeviceName
					);

	if( pDeviceLink->priv )
	{
		kfree(pDeviceLink->priv);
	}

	if( pDeviceExtension->pucDeviceName )
	{
		kfree( pDeviceExtension->pucDeviceName );
		pDeviceExtension->pucDeviceName = NULL;
	}

	if( pDeviceExtension )
	{
		kfree ( pDeviceExtension );
		pDeviceExtension = NULL;
	}
	DebugLog("Driver_Detach Exit\n");
}

void cs_error(client_handle_t handle, int func, int ret)
{
/*=======================================================
Used to report errors to Card Services
=========================================================*/
    error_info_t err = { func, ret };
    CardServices(ReportError, handle, &err);
}

void Driver_Release(u_long arg)
{
	// After a card is removed, SCR241_release() will unregister the
	// device, and release the PCMCIA configuration.  If the device is
	// still open, this will be postponed until it is closed.

	dev_link_t *pDeviceLink = (dev_link_t *)arg;

	// If the device is currently in use, we won't release until it
	// is actually closed, because until then, we can't be sure that
	// no one will try to access the device or its data structures.
	if (pDeviceLink->open)
	{
		pDeviceLink->state |= DEV_STALE_CONFIG;
		return;
	}

	// Unlink the device chain
	pDeviceLink->dev = NULL;


	// In a normal driver, additional code may be needed to release
	// other kernel data structures associated with this device.

	// Don't bother checking to see if these succeed or not
	if (pDeviceLink->win)
	{
		CardServices(ReleaseWindow, pDeviceLink->win);
	}
	
	CardServices(ReleaseConfiguration, pDeviceLink->handle);
	if (pDeviceLink->io.NumPorts1)
	{
		CardServices(ReleaseIO, pDeviceLink->handle, &pDeviceLink->io);
	}
	if (pDeviceLink->irq.AssignedIRQ)
	{
		// we verify if the IRQ is already released by the DFU
		// operation. if not, only then we release it here
		CardServices(ReleaseIRQ, pDeviceLink->handle, &pDeviceLink->irq);
	}
	pDeviceLink->state &= ~DEV_CONFIG;

	if (pDeviceLink->state & DEV_STALE_LINK)
	{
		DebugLog("Detaching from Driver_Release \n");
		Driver_Detach(pDeviceLink);
	}
	DebugLog("Driver_Release Exit\n");
}

int Event_Handler(
	event_t event, int priority,
	event_callback_args_t *args
	)
{
	/*======================================================================
	The card status event handler.  Mostly, this schedules other
	stuff to run after an event is received.

	When a CARD_REMOVAL event is received, we immediately set a
	private flag to block future accesses to this device.  All the
	functions that actually access the device should check this flag
	to make sure the card is still present.
	======================================================================*/
	dev_link_t *pDeviceLink = args->client_data;
	int ncsRetVal=0;

	switch (event) {
	case CS_EVENT_CARD_REMOVAL:
		pDeviceLink->state &= ~DEV_PRESENT;

		if (pDeviceLink->state & DEV_CONFIG)
		{
			pDeviceExtension->nStop = 1;
			mod_timer(&pDeviceLink->release, 10000 ); //jiffies + HZ/20
		}
		break;
	case CS_EVENT_CARD_INSERTION:
	
		pDeviceLink->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
		pDeviceExtension->bus = args->bus;
		Driver_Config(pDeviceLink);
		pDeviceExtension->nStop = 0;
		break;
	case CS_EVENT_PM_SUSPEND:
		pDeviceLink->state |= DEV_SUSPEND;
		// Fall through...
	case CS_EVENT_RESET_PHYSICAL:
		// Mark the device as stopped, to block IO until later
		pDeviceExtension->nStop = 1;
		if (pDeviceLink->state & DEV_CONFIG)
		{
			CardServices(ReleaseConfiguration, pDeviceLink->handle);
		}
		break;
	case CS_EVENT_PM_RESUME:
		DebugLog("<1> Evnet PM_RESUME Called \n");
		pDeviceLink->state &= ~DEV_SUSPEND;
		
		//
		//If the Irq was not assigned or if polled mode has been set 
		//then start the thread explicitly on resume
		//	
		if(FALSE == pDeviceExtension->bIsIrqSet  ||TRUE == pDeviceExtension->bIsPolledMode)
		{
			DebugLog("<1>******* THREAD TO BE STARTED ON RESUME*****\n");
			start_kthread(
						PcdmCardStatusThread,
						&pDeviceExtension->CardTracker );
		}
			
		/* Fall through... */
	case CS_EVENT_CARD_RESET:
		if (pDeviceLink->state & DEV_CONFIG)
		{
			CardServices(
						RequestConfiguration,
						pDeviceLink->handle,
						&pDeviceLink->conf
						);
		}
		
		// since the device went to standby/suspend, the firmware
		// will get reset to Rev A status. Hence, we need to update
		// the transport mechanism used prior to suspend/standby to the firmware.
		if( CS_EVENT_PM_RESUME == event )
		{
			
			if( PCDM_REV_C == pDeviceExtension->byPcdmRev )
			{
				DebugLog("<1>*********** Found REVC ON RESUME***************\n");
				pDeviceExtension->byPcdmRev = PCDM_REV_A;
				ncsRetVal = IFDSetTransportMechanism( pDeviceExtension );
				if( 0 == ncsRetVal )
				{
					pDeviceExtension->byPcdmRev = PCDM_REV_C;
				}
				else
				{
					// we failed to the tranport mechanism. hence, we revert
					// back to rev-A. we dont take into account the return
					// state 
					pDeviceExtension->byPcdmRev = PCDM_REV_A;
					ncsRetVal = IFDSetTransportMechanism( pDeviceExtension );
					DebugLog(\
					"<1>Failure in Setting transport Mechanism to Rev-C. Revert to Rev-A\n" 
					);
					break;
				}
			}
			IFDGetCardStatus(pDeviceExtension);
			pDeviceExtension->bIsSystemResume = TRUE;
		}
		//In a normal driver, additional code may go here to restore
		//the device state and restart IO.
		pDeviceExtension->nStop = 0;
		break;
	}
	return ncsRetVal;
} /* SCR241_event */

static void Driver_Config( dev_link_t *link )
{
	/*======================================================================
	This function parses through the CIS and prepares the card for IO
	operations.
	======================================================================*/
	client_handle_t handle 			= link->handle;
	tuple_t 		tuple;
	cisparse_t 		parse;
	int 			last_fn,
					last_ret;
	u_char 			buf[64];
	config_info_t 	conf;
	win_req_t 		req;
	memreq_t 		map;
	cisinfo_t 		cisinfo;
	int 			ret=0;
	int 		instanceCount =0;
	int				nRetVal;
//	ds_ioctl_arg_t	arg;
//	cistpl_manfid_t	manfid;


	DebugLog("<1> Configuring the module \n");

	// This reads the card's CONFIG tuple to find its
	// configuration registers.
	ret = CardServices(ValidateCIS,handle,&cisinfo);
	tuple.DesiredTuple 	= CISTPL_CONFIG;
	tuple.Attributes 	= 0;
	tuple.TupleData 	= buf;
	tuple.TupleDataMax 	= sizeof(buf);
	tuple.TupleOffset 	= 0;

	CS_CHECK(GetFirstTuple, handle, &tuple);
	CS_CHECK(GetTupleData, handle, &tuple);
	CS_CHECK(ParseTuple, handle, &tuple, &parse);

	link->conf.ConfigBase = parse.config.base;
	link->conf.Present = parse.config.rmask[0];

	// Configure card
	link->state |= DEV_CONFIG;

	// Look up the current Vcc
	CS_CHECK(GetConfigurationInfo, handle, &conf);
	link->conf.Vcc = conf.Vcc;

	/*
	In this loop, we scan the CIS for configuration table entries,
	each of which describes a valid card configuration, including
	voltage, IO window, memory window, and interrupt settings.

	We make no assumptions about the card to be configured: we use
	just the information available in the CIS.  In an ideal world,
	this would work for any PCMCIA card, but it requires a complete
	and accurate CIS.  In practice, a driver usually "knows" most of
	these things without consulting the CIS, and most client drivers
	will only use the CIS to fill in implementation-defined details.
	*/
	tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
	CS_CHECK(GetFirstTuple, handle, &tuple);
	while (1)
	{
		cistpl_cftable_entry_t dflt = { 0 };
		cistpl_cftable_entry_t *cfg = &(parse.cftable_entry);

		CFG_CHECK(GetTupleData, handle, &tuple);
		CFG_CHECK(ParseTuple, handle, &tuple, &parse);

		if (cfg->flags & CISTPL_CFTABLE_DEFAULT)
		{
			dflt = *cfg;
		}
		if (cfg->index == 0)
		{
			goto next_entry;
		}
		link->conf.ConfigIndex = cfg->index;

		// Does this card need audio output?
		if (cfg->flags & CISTPL_CFTABLE_AUDIO)
		{
			link->conf.Attributes |= CONF_ENABLE_SPKR;
			link->conf.Status = CCSR_AUDIO_ENA;
		}
		
		// Use power settings for Vcc and Vpp if present
		// Note that the CIS values need to be rescaled
		if (cfg->vcc.present & (1<<CISTPL_POWER_VNOM))
		{
			if (conf.Vcc != cfg->vcc.param[CISTPL_POWER_VNOM]/10000)
			{
				goto next_entry;
			}
		}
		else if (dflt.vcc.present & (1<<CISTPL_POWER_VNOM))
		{
			if (conf.Vcc != dflt.vcc.param[CISTPL_POWER_VNOM]/10000)
			{
				goto next_entry;
			}
		}

		if (cfg->vpp1.present & (1<<CISTPL_POWER_VNOM))
		{
			link->conf.Vpp1 = link->conf.Vpp2 =
				cfg->vpp1.param[CISTPL_POWER_VNOM]/10000;
		}
		else if (dflt.vpp1.present & (1<<CISTPL_POWER_VNOM))
		{
			link->conf.Vpp1 = link->conf.Vpp2 =
				dflt.vpp1.param[CISTPL_POWER_VNOM]/10000;
		}

		if (cfg->irq.IRQInfo1 || dflt.irq.IRQInfo1)
		{
			 link->conf.Attributes |= CONF_ENABLE_IRQ;
		}

		// IO window settings
		link->io.NumPorts1 = link->io.NumPorts2 = 0;
		if ((cfg->io.nwin > 0) || (dflt.io.nwin > 0))
		{
			cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt.io;
			link->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
			if (!(io->flags & CISTPL_IO_8BIT))
			{
				link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
				DebugLog(KERN_INFO "IO_DATA_PATH_WIDTH_16\n");
			}
			if (!(io->flags & CISTPL_IO_16BIT))
			{
				link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
				DebugLog(KERN_INFO "IO_DATA_PATH_WIDTH_8\n");
			}
			link->io.IOAddrLines = io->flags & CISTPL_IO_LINES_MASK;
			link->io.BasePort1 = io->win[0].base;
			link->io.NumPorts1 = io->win[0].len;
			if (io->nwin > 1)
			{
				link->io.Attributes2 = link->io.Attributes1;
				link->io.BasePort2 = io->win[1].base;
				link->io.NumPorts2 = io->win[1].len;
			}
			// This reserves IO space but doesn't actually enable it
			CFG_CHECK(RequestIO, link->handle, &link->io);
		}
		/*
		Now set up a common memory window, if needed.  There is room
		in the dev_link_t structure for one memory window handle,
		but if the base addresses need to be saved, or if multiple
		windows are needed, the info should go in the private data
		structure for this device.

		Note that the memory window base is a physical address, and
		needs to be mapped to virtual space with ioremap() before it
		is used.
		*/
		if ((cfg->mem.nwin > 0) || (dflt.mem.nwin > 0))
		{
			cistpl_mem_t *mem =
				(cfg->mem.nwin) ? &cfg->mem : &dflt.mem;
			req.Attributes = WIN_DATA_WIDTH_16|WIN_MEMORY_TYPE_CM;
			req.Attributes |= WIN_ENABLE;
			req.Base = mem->win[0].host_addr;
			req.Size = mem->win[0].len;
			if (req.Size < 0x1000)
			{
				req.Size = 0x1000;
			}
			req.AccessSpeed = 0;
			link->win = (window_handle_t)link->handle;
			CFG_CHECK(RequestWindow, &link->win, &req);
			map.Page = 0; map.CardOffset = mem->win[0].card_addr;
			CFG_CHECK(MapMemPage, link->win, &map);
		}

		break;

	next_entry:
		if (link->io.NumPorts1)
		CardServices(ReleaseIO, link->handle, &link->io);
		CS_CHECK(GetNextTuple, handle, &tuple);
	}

	// Allocate an interrupt line.  Note that this does not assign a
	// handler to the interrupt, unless the 'Handler' member of the
	// irq structure is initialized.
	if (link->conf.Attributes & CONF_ENABLE_IRQ)
	{
		nRetVal = CardServices(RequestIRQ, link->handle, &link->irq);
		if( nRetVal == CS_SUCCESS )
		{
			DebugLog("<1> Interrupt Successfully Assigned \n");
			pDeviceExtension->bIsIrqSet = TRUE;
			pDeviceExtension->bOrgIrqStatus = TRUE;
		}
		else
		{
			pDeviceExtension->bIsIrqSet = FALSE;
			pDeviceExtension->bOrgIrqStatus = FALSE;
		}
		if( nRetVal & CS_IN_USE )
		{
			DebugLog(
					"<1> Requested Interupt not available, \
					or already assigned \n"
					);
		}
		if( nRetVal & CS_BAD_ATTRIBUTE )
		{
			DebugLog("<1> BAD ATTRIBUTE SET \n");
		}
		if( nRetVal & CS_CONFIGURATION_LOCKED )
		{
			DebugLog(" Driver_SetIrq Entry !!! \n");
			DebugLog("<1> CONFIGURATION IS ALREADY LOCKED \n");
		}

		if( link->irq.IRQInfo1 & IRQ_INFO2_VALID )
		{
			if( link->irq.IRQInfo2 & IRQ_HANDLE_PRESENT )
			{
				DebugLog(
						"<1> IRQ-HANDLE-PRESENT Macro is set. \
						Handler is being used . \n"
						);
			}
		}
	}
	// DebugLog("<1> INTERRUPT SUCCESSFULLY ASSIGNED \n");
	// This actually configures the PCMCIA socket -- setting up
	// the I/O windows and the interrupt mapping, and putting the
	// card and host interface into "Memory and IO" mode.
	CS_CHECK(RequestConfiguration, link->handle, &link->conf);

	// We can release the IO port allocations here, if some other
	// driver for the card is going to loaded, and will expect the
	// ports to be available.
	if (free_ports)
	{
		if (link->io.BasePort1)
		{
			release_region(link->io.BasePort1, link->io.NumPorts1);
		}
		if (link->io.BasePort2)
		{
			release_region(link->io.BasePort2, link->io.NumPorts2);
		}
	}

	// At this point, the dev_node_t structure(s) need to be
	// initialized and arranged in a linked list at link->dev.
	sprintf(pDeviceExtension->DeviceNode.dev_name, "SCR24x%d",instanceCount++);
	pDeviceExtension->DeviceNode.major = majorn;/*SCR241_MAJ_VER;*/
	pDeviceExtension->DeviceNode.minor = 0;

	// Maps Physical Memmory To Local Data Structure For IO
	// Using physical Base requires (caddr_t) cast
	pDeviceExtension->am_base = ioremap(req.Base, req.Size);
	pDeviceExtension->IoBase = link->io.BasePort1;

	// Make Device Link
	link->dev = &pDeviceExtension->DeviceNode;

	// Finally, report what we've done
	// dev->node.dev_name, link->conf.ConfigIndex,
	// link->conf.Vcc/10, link->conf.Vcc%10);
	if (link->conf.Vpp1)
	{
		DebugLog(
				", Vpp %d.%d",
				link->conf.Vpp1/10,
				link->conf.Vpp1%10
				);
	}
	if (link->conf.Attributes & CONF_ENABLE_IRQ)
	{
		DebugLog(
				", irq %d", \
				((int)pDeviceExtension->nAssignedIRQ = \
				link->irq.AssignedIRQ)
				);
	}
	if (link->io.NumPorts1)
	{
		DebugLog(", io 0x%04x-0x%04x", link->io.BasePort1,
		link->io.BasePort1+link->io.NumPorts1-1);
	}
	if (link->io.NumPorts2)
	{
		DebugLog(" & 0x%04x-0x%04x", link->io.BasePort2,
		link->io.BasePort2+link->io.NumPorts2-1);
	}
	if (link->win)
	{
		DebugLog(
				", mem 0x%06lx-0x%06lx",
				req.Base,
				req.Base+req.Size-1
				);
	}
	DebugLog("\n");
	link->state &= ~DEV_CONFIG_PENDING;

	Driver_SetIOBase(pDeviceExtension,link->io.BasePort1);

	pDeviceExtension->bSignalEvent = FALSE;

	nRetVal = Driver_InitLocal(pDeviceExtension);
	pDeviceExtension->irqFlag = FALSE;
	pDeviceExtension->bIsQueueSet = FALSE;
	
	if( STATUS_SUCCESS != nRetVal )
	{
		DebugLog("<1> Device is NOT Present \n");
		return;
	}
	
	if( FALSE == pDeviceExtension->bIsIrqSet )
	{
		//start_kthread( PcdmCardStatusThread, &pDeviceExtension->CardTracker );
		//pDeviceExtension->bIsPolledMode = TRUE;
	}
	return;

	cs_failed:
		DebugLog("<1> Failure has occured in allocation !!!! \n");
		cs_error(link->handle, last_fn, last_ret);
		Driver_Release((u_long)link);

} // SCR241_config

void Driver_SetIOBase(DEVICE_EXTENSION *pDeviceExtension,u_long iobase)
{
	//====================================================================
	//			copy the iobase to the device extension
	//====================================================================
	pDeviceExtension->IoBase = iobase;

}

NTSTATUS Driver_InitLocal(DEVICE_EXTENSION  *pDeviceExtension)
{
	NTSTATUS NTStatus = STATUS_SUCCESS;
	//====================================================================
	//		Initialization of the device extension struction
	//====================================================================
	DebugLog ("<1> SCR241:: Driver_InitLocal Entry of the Kernel Module\n");

	do
	{
		pDeviceExtension->irqFlag		 	= 0;
		pDeviceExtension->inIOCTL 			= 0;
		pDeviceExtension->prevCardStatus 	= 0;
		pDeviceExtension->cardStsChange		= FALSE;
		pDeviceExtension->cmd_lock 			= SPIN_LOCK_UNLOCKED;
		pDeviceExtension->ioctl_lock		= SPIN_LOCK_UNLOCKED;
		pDeviceExtension->rdwr_lock			= SPIN_LOCK_UNLOCKED;

		spin_lock_init(&pDeviceExtension->cmd_lock);
		spin_lock_init(&pDeviceExtension->ioctl_lock);
		spin_lock_init(&pDeviceExtension->rdwr_lock);

		init_waitqueue_head(&pDeviceExtension->wait);
		init_waitqueue_head(&pDeviceExtension->delay);
		
		NTStatus = IFDGetInterfaceDetails( pDeviceExtension );
		if( STATUS_SUCCESS != NTStatus )
		{
			pDeviceExtension->bBootRomPresent = TRUE;
			DebugLog("<1> the device is in BOOT ROM MODE \n");
			break;
		}
		NTStatus = IFDGetCardStatus(pDeviceExtension);
		DebugLog (
				"<1> SCR241:: Driver_InitLocal Exit of the Kernel Module [%ld]\n",
				NTStatus
				);
	}while(FALSE);
	return NTStatus;
}

void CardStatusCheck(void  *arg)
{
	//====================================================================
	//			Checks the presence of the smart card
	//=====================================================================
	DEVICE_EXTENSION	*pDeviceExtension =  *(DEVICE_EXTENSION **)arg;

	DebugLog ("<1>SCR241:: CardStatusCheck Entry of Kernel Module\n");
	spin_lock(&pDeviceExtension->cmd_lock);
	pDeviceExtension->irqFlag = FALSE;


	IFDGetCardStatus(pDeviceExtension);
	// added by joe
	if( TRUE == pDeviceExtension->bNoIRQ )
	{
		queue_task(&pDeviceExtension->tqueue ,&tq_scr241);
	}
	spin_unlock(&pDeviceExtension->cmd_lock);

}

int IFDHOpen(struct inode *inode , struct file *file )
{
	//======================================================================
	// 	This function is called when the IfdHandler opens the device.
	//=======================================================================
	int minor = MINOR(inode->i_rdev);
	dev_link_t *link;

	DebugLog ("<1> Calling IFDHOpen of the KERNEL MODULE OF SCR241\n");

	for (link = dev_list; link; link = link->next)
	{
		if (link->dev && link->dev->minor == minor)
		{
			break;
		}
	}

	if (!DEV_OK(link))
	{
		return -ENODEV;
	}

	if( TRUE == pDeviceExtension->bNoIRQ )
	{
		//run_task_queue(&tq_scr241);
	}

	pDeviceExtension->cardStsChange=TRUE;
	MOD_INC_USE_COUNT;

	return 0;
}


int IFDHClose(struct inode *inode , struct file *file )
{
	//======================================================================
	//	Called when the Ifdhandler closes communication with the device
	//======================================================================
	DebugLog ("<1> Calling IFDHClose of the KERNEL MODULE OF SCR241\n");
	do
	{
		if( NULL == pDeviceExtension )
		{
			break;
		}
		if( TRUE == pDeviceExtension->bIsPolledMode )
		{
			DebugLog(
					"<1> Killing Polled Mode Card Tracking from IFDHClose %d\n", 
					pDeviceExtension->bOrgIrqStatus 
					);
			
			pDeviceExtension->bIsPolledMode = FALSE;
			pDeviceExtension->bIsIrqSet = pDeviceExtension->bOrgIrqStatus;

			if( TRUE == pDeviceExtension->bOrgIrqStatus )
			{
				/* Kill the card tracking thread */
				DebugLog(" STOPPING CARD TRACKING THREAD AS IRQ is absent \n" );
				stop_kthread( &pDeviceExtension->CardTracker );
			}
		}
	}while( FALSE ); 
	MOD_DEC_USE_COUNT;
	return 0;
}


int IFDHIoctl(struct inode *inode, struct file *file, u_int ucIoctl, u_long arg)
{
	//======================================================================
	//				ioctl calls from the ifdhandler.
	//				cmd is the IOCTL code
	//				arg is the ptr to in\out buffer
	//======================================================================
	int 			minor 		= MINOR(inode->i_rdev);
	dev_link_t 		*link;
	u_char			*buf;
	PIOCTL_DATA 	pIoctlData;
	PTXRX_DATA		pTxRxData;
	BYTE			byStatus 	= 0;
	int 			nCount;
	int 			nRetVal		= 0;
	DWORD 			dwLen;
	PUCHAR			pucTemp;
	ULONG			ulSentLen,
				ulOffset;
	buf 		= (u_char *)arg;
	pIoctlData 	= (PIOCTL_DATA) buf;
	pTxRxData 	= (PTXRX_DATA) buf;

	// Determine Recipient of IOCTL
	for (link = dev_list; link; link = link->next)
	{
		if (link->dev && link->dev->minor == minor)
		{
			break;
		}
	}

	if (!DEV_OK(link))
	{
		DebugLog ("Failed in Link\n");
		return -ENODEV;
	}
	
	if( NULL == pDeviceExtension )
	{
		return -ENODEV;
	}

	if(pDeviceExtension->irqFlag == TRUE)
	{
		pDeviceExtension->irqFlag = FALSE;
		//run_task_queue(&tq_scr241);
	}

	pucTemp = (PUCHAR) kmalloc(sizeof(UCHAR)*2 , GFP_KERNEL);
	if( NULL == pucTemp)
	{
		DebugLog("Error Allocating Memory\n");
		return -ENODEV;
	}

	switch ( ucIoctl )
	{
		case CMD_CARD_STATUS:
			pucTemp[0] = pDeviceExtension->byCardStatus;
			copy_to_user(
				pIoctlData->ucRxBuffer,
				(PUCHAR) pucTemp,
				1 );

			pIoctlData->dwRxLength = 2;
			byStatus = 2;
			nRetVal = 2;
			break;
			
		case CMD_BOOTROM_STATUS:
			pucTemp[0] = pDeviceExtension->bBootRomPresent;
			copy_to_user(
				pIoctlData->ucRxBuffer,
				(PUCHAR) pucTemp,
				1 );
			byStatus = 2;
			nRetVal = 2;
			break;

		case CMD_PCCARD_DOWNLOAD_START:
			byStatus = IFDPCCardDownloadStart (
				pDeviceExtension , buf );
			if( STATUS_SUCCESS != byStatus )
			{
				byStatus = -ENODEV;
			}
			pDeviceExtension->pucFwBuffer = (PUCHAR) 
				kmalloc(sizeof(UCHAR)*65536 , GFP_KERNEL);
				
			if( NULL ==  pDeviceExtension->pucFwBuffer )
			{
				byStatus = -ENODEV;
				break;
			}
			memset( pDeviceExtension->pucFwBuffer , 0, 65536 );
			pDeviceExtension->ulOffset = 0;
			nRetVal = 1;
			break;
			
		case CMD_PCCARD_DOWNLOAD_DATA:
			ulSentLen = pIoctlData->dwTxLength;
			ulOffset = pDeviceExtension->ulOffset;
/*			copy_from_user( 
				(PUCHAR)(pDeviceExtension->pucFwBuffer+ulOffset ),
				(PUCHAR) pIoctlData->ucRxBuffer,
				(ULONG) ulSentLen
				);
*/
			memcpy( (PUCHAR)pDeviceExtension->pucFwBuffer+ulOffset, (PUCHAR)pIoctlData->ucTxBuffer,ulSentLen); 
			for(nCount=0;nCount< ulSentLen; nCount++ )
			{
				//DebugLog("<1> <(%02x)>, ", (pDeviceExtension->pucFwBuffer[ulOffset+nCount] ) );
			}
			pDeviceExtension->ulOffset += ulSentLen;
			nRetVal = 1;
			break;
			
		case CMD_PCCARD_DOWNLOAD_END:
			byStatus = IFDPerformPCCardDownLoad(
				pDeviceExtension , pDeviceExtension->pucFwBuffer );
			if( STATUS_SUCCESS != byStatus )
			{
				byStatus = -ENODEV;
				break;
			}
			nRetVal = 1;
			//pIoctlData->dwRxLength = 2;
			break;

		case CMD_PCCARD_ERASE:
			IFDPerformPCCardErase( pDeviceExtension );
			break;

		case CMD_WRITE_DATA:
			DebugLog( " Writing to Device \n");
			if( TRUE == pDeviceExtension->bBootRomPresent )
			{
				break;
			}
			nRetVal = LocalWrite(
							pIoctlData->ucTxBuffer,
							pIoctlData->dwTxLength );
			break;

		case CMD_READ_DATA:
			DebugLog( "<1> Reading to Device \n");
			if( TRUE == pDeviceExtension->bBootRomPresent )
			{
				break;
			}
			dwLen = LocalRead(
							pIoctlData->ucRxBuffer,
							pIoctlData->dwRxLength );
			if( dwLen > 0 )
			{
				byStatus = dwLen;
				nRetVal = (int) dwLen;
			}
			pIoctlData->dwRxLength = dwLen;
			DebugLog(
				"<1> Reading Data of Length [%ld] \n",
				pIoctlData->dwRxLength );

			for(nCount=0; nCount < pIoctlData->dwRxLength; nCount++ )
			{
				DebugLog(" [%02X] ", pIoctlData->ucRxBuffer[nCount] );
			}
			break;
		case CMD_POLLED_MODE:
			DebugLog(
				" IFDHIOCTL :: CMD_POLLED_MODE is %02X \n",
				pIoctlData->ucRxBuffer[0]
				);

			if( 0x01 == pIoctlData->ucRxBuffer[0] )
			{
				if(FALSE == pDeviceExtension->bIsPolledMode ) 
				{
					pDeviceExtension->bIsPolledMode = TRUE;

					start_kthread(
						PcdmCardStatusThread,
						&pDeviceExtension->CardTracker );
				}
			}
			else
			{
				if( TRUE == pDeviceExtension->bIsPolledMode )
				{
					DebugLog("<1> Killing Polled Mode Card Tracking \n");
					pDeviceExtension->bIsPolledMode = FALSE;
					pDeviceExtension->bIsIrqSet = pDeviceExtension->bOrgIrqStatus;
					/* Kill the card tracking thread */
					stop_kthread( &pDeviceExtension->CardTracker );
				}
			}
			break;
		case CMD_GET_IRQSTATUS:
			pucTemp[0] = pDeviceExtension->bIsIrqSet;
			copy_to_user(
			pIoctlData->ucRxBuffer,
				pucTemp,
				1 );
			break;

		case CMD_SET_REVISION:
			DebugLog("<1>Setting PCDM Revision \n");
			copy_from_user(
				pucTemp,
				(PUCHAR) pIoctlData->ucRxBuffer,
				(DWORD) 1
				);
			pDeviceExtension->byPcdmRev = pucTemp[0];
			break;

		default:
			DebugLog("<0> WRONG IOCTL being sent %d \n", ucIoctl ); 
			byStatus = -ENODEV;
			nRetVal = -ENODEV;
			break;
	}
	pDeviceExtension->inIOCTL = FALSE;
	if( pucTemp )
	{
		kfree(pucTemp);
		pucTemp = NULL;
	}
	return nRetVal; //TBD  to return appropriate values.
}

ssize_t IFDHRead(
	struct file 	*file,
	char 			*pucRxBuffer,
	size_t 			count,
	loff_t 			*loc
	)
{
	PIOCTL_DATA pIoctlData = (PIOCTL_DATA) pucRxBuffer;
	if( NULL != pDeviceExtension )
	{
		if( TRUE == pDeviceExtension->bIsSystemResume )
		{
		
			pIoctlData->ucRxBuffer[0] = READER_STS_CARD_ABSENT;
			pDeviceExtension->bIsSystemResume = FALSE;
		}
		else
		{
			pIoctlData->ucRxBuffer[0] = pDeviceExtension->byCardStatus;
		}
		pIoctlData->dwRxLength = 1;
		return 1;
	}
	else
	{
		pIoctlData->ucRxBuffer[0] = READER_STS_CARD_ABSENT;
		pIoctlData->dwRxLength = 1;
		return 1;
	}
}
DWORD LocalRead( char *pucRxBuffer, size_t count )
{
	/*======================================================================

	ioctl calls from the ifdhandler.
	cmd is the IOCTL code
	arg is the ptr to in\out buffer

	======================================================================*/
	PUCHAR pucData = NULL;
	NTSTATUS NTStatus = STATUS_SUCCESS;
	DWORD dwDataLen;
	int i;

	DebugLog ("<1> Calling IFDHREAD of the KERNEL MODULE OF SCR241 n");

	pucData = (PUCHAR) kmalloc(
		(sizeof(UCHAR) * MAX_DATA_SIZE) ,
		GFP_KERNEL
		);

	if( NULL == pucData )
	{
		return -ENODEV; // TBD
	}
	memset ( pucData, 0, ( sizeof(UCHAR)*MAX_DATA_SIZE) );

	NTStatus = PcdmRead (
		pDeviceExtension,
		pucData,
		&dwDataLen
		);

	if( STATUS_SUCCESS == NTStatus )
	{
		/*for(i=0;i < dwDataLen; i++ )
		{
			DebugLog(
				"<1> IFDHREAD::Before Copying, Recieved Values %02X \n",
				*(pucData+i));
		}*/
		copy_to_user( pucRxBuffer, pucData, dwDataLen);
		for(i=0;i < dwDataLen; i++ )
		{
			DebugLog("<1> (%02X) \n", *(pucRxBuffer+i));
		}
		if( pucData )
		{
			kfree( pucData );
			pucData = NULL;
		}
		return dwDataLen;
	}
	else
	{
		if( pucData )
		{
			kfree( pucData );
			pucData = NULL;
		}
		return 0;
	}


}

ssize_t IFDHWrite(
	struct file *file,
	const char *pucTxBuffer,
	size_t szCount,
	loff_t *loc)
{
	return -ENODEV;
}

ssize_t LocalWrite( const char *pucTxBuffer, size_t szCount)
{
	PUCHAR 		pucData = NULL;
	NTSTATUS 	NTStatus = STATUS_SUCCESS;
	DWORD 		dwDataLen = 0;
	int 		nI;

	DebugLog ("Calling IFDHWRITE of the KERNEL MODULE OF SCR241 \n");

	pucData = (PUCHAR) kmalloc( (sizeof(UCHAR) * 2048) , GFP_KERNEL );
	if( NULL == pucData )
	{
		return -ENODEV;
	}
	memset ( pucData, 0, ( sizeof(UCHAR)*2048) );

	copy_from_user( (PUCHAR)pucData, (PUCHAR)pucTxBuffer, (DWORD)szCount );
	DebugLog("Length of data being sent down from IFDHWrite is %d", szCount);
	for(nI=0;nI<(int)szCount;nI++)
	{
		DebugLog("<1>Writing--> %02X\n", *(pucData+nI));
	}
	dwDataLen = (DWORD)szCount;

	NTStatus = PcdmWrite (
		pDeviceExtension,
		pucData,
		dwDataLen
		);

	if( STATUS_SUCCESS == NTStatus )
	{
		if( pucData )
		{
			kfree( pucData );
			pucData = NULL;
		}
		return szCount;
	}
	else
	{
		if( pucData )
		{
			kfree( pucData );
			pucData = NULL;
		}
		return 0;
	}

}

void Driver_IrqHandler(int irq,void *pDev,struct pt_regs *regs)
{
	/*====================================================================
	This is the Interrupt Handler
	Interrupts occur during Smart Card Insertion and removal.
	Hence card status is updated here.
	IFDGetCardStatus() sends a command to the reader, and command can be sent
	only if there is no other command in progress in the reader.
	If a command is in progress, the task is queued
	=====================================================================*/
	DEVICE_EXTENSION *pDeviceExtension;
	pDeviceExtension = pDev;

	//DebugLog( "<1> Calling Driver_IrqHandler\n" );

	// we dont use the interuppt handler if forced polled
	// mode is set in the registry.
	if( TRUE == pDeviceExtension->bIsPolledMode )
	{
		DebugLog("<1> Polling is set. Hence, not handled here\n");
		//return;
	}

	//DebugLog ("Calling Interrupt Mode Card Tracking in KERNEL MODULE \n");
	PcmciaIrqServiceRoutine ( pDeviceExtension );
	return;
}

void printdummy(char *dbg_msg, ...)
{
	// This is a dummy function to override the debug data
}

void tqGetCardStatus( void *arg)
{
	/*==========================================================
	This is a scheduler queue used to trigger the card status
	function if the interrupt got is from our device
	===========================================================*/
	PDEVICE_EXTENSION pDeviceExtension;
	pDeviceExtension = (PDEVICE_EXTENSION) arg;
	if( FALSE == pDeviceExtension->bBootRomPresent )
	{
		IFDGetCardStatus( pDeviceExtension );
	}
	pDeviceExtension->bIsQueueSet = FALSE;
}

module_init( init_driver );
module_exit( exit_driver );

MODULE_LICENSE ("GPL");
